@@ -41,8 +41,8 @@
 #include <libgnomevfs/gnome-vfs-utils.h>
 #include <math.h>
 #include <stdlib.h>
+#include <stdio.h>
 #include <pango/pangoft2.h>
-#include <png.h>
 
 #define LOAD_BUFFER_SIZE 65536
 
