@@ -694,7 +694,7 @@ eel_gnome_icon_selector_new (const char *title,
 			     EelIconSelectionFunction selected,
 			     gpointer callback_data)
 {
-	GtkWidget *dialog, *icon_selection;
+	GtkWidget *dialog, *icon_selection, *retval;
 	GtkWidget *entry, *file_entry;
 	IconSelectionData *selection_data;
 	
@@ -756,10 +756,14 @@ eel_gnome_icon_selector_new (const char *title,
 
 	gnome_icon_selection_show_icons (GNOME_ICON_SELECTION (icon_selection));
 
+	/* eel_remove_weak_pointer() will set dialog to NULL, so we need to
+	   remember the dialog pointer here, if there is one. */
+	retval = dialog;
+
 	/* Now remove the weak pointer, if the dialog still exists. */
 	eel_remove_weak_pointer (&dialog);
 
-	return dialog;
+	return retval;
 }
 
 PangoContext *
