@@ -48,6 +48,7 @@
 #include <gtk/gtkseparatormenuitem.h>
 #include <gtk/gtksignal.h>
 #include <pango/pangoft2.h>
+#include <math.h>
 
 #include "eel-marshal.h"
 #include "eel-marshal.c"
@@ -80,12 +81,17 @@
 void
 eel_gtk_button_set_padding (GtkButton *button, int pad_amount)
 {
+	GtkWidget *child;
+
 	g_return_if_fail (GTK_IS_BUTTON (button));
 	g_return_if_fail (pad_amount > 0);
 
-	gtk_misc_set_padding (GTK_MISC (GTK_BIN(button)->child), 
-			      pad_amount, 
-			      pad_amount);
+	child = GTK_BIN (button)->child;
+
+	if (GTK_IS_MISC (child)) {
+		gtk_misc_set_padding
+			(GTK_MISC (child), pad_amount, pad_amount);
+	}
 }
 
 /**
@@ -497,9 +503,14 @@ eel_pop_up_context_menu (GtkMenu	     *menu,
 	 * other than the one that invoked the menu will be ignored (instead
 	 * of dismissing the menu). This is a subtle fragility of the GTK menu code.
 	 */
-	button = event->type == GDK_BUTTON_RELEASE
-		? 0
-		: event->button;
+
+	if (event) {
+		button = event->type == GDK_BUTTON_RELEASE
+			? 0
+			: event->button;
+	} else {
+		button = 0;
+	}
 	
 	gtk_menu_popup (menu,					/* menu */
 			NULL,					/* parent_menu_shell */
@@ -507,7 +518,7 @@ eel_pop_up_context_menu (GtkMenu	     *menu,
 			eel_popup_menu_position_func,	/* func */
 			&offset,			        /* data */
 			button,					/* button */
-			event->time);				/* activate_time */
+			event ? event->time : GDK_CURRENT_TIME); /* activate_time */
 
 	gtk_object_sink (GTK_OBJECT (menu));
 }
@@ -934,6 +945,141 @@ eel_gtk_label_set_scale (GtkLabel *label,
 	pango_attr_list_unref (attr_list);
 }
 
+static void
+get_layout_location (GtkLabel  *label,
+                     gint      *xp,
+                     gint      *yp)
+{
+  GtkMisc *misc;
+  GtkWidget *widget;
+  float xalign;
+  int x, y;
+  int shadow_offset;
+  
+  shadow_offset = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (label),
+						      "eel-label-shadow-offset"));
+  
+  misc = GTK_MISC (label);
+  widget = GTK_WIDGET (label);
+  
+  if (gtk_widget_get_direction (widget) == GTK_TEXT_DIR_LTR)
+    xalign = misc->xalign;
+  else
+    xalign = 1.0 - misc->xalign;
+  
+  x = floor (widget->allocation.x + (int)misc->xpad
+             + ((widget->allocation.width - widget->requisition.width - shadow_offset) * xalign)
+             + 0.5);
+  
+  y = floor (widget->allocation.y + (int)misc->ypad 
+             + ((widget->allocation.height - widget->requisition.height - shadow_offset) * misc->yalign)
+             + 0.5);
+  
+
+  if (xp)
+    *xp = x;
+
+  if (yp)
+    *yp = y;
+}
+
+static gboolean
+eel_gtk_label_expose_event (GtkLabel *label, GdkEventExpose *event, gpointer user_data)
+{
+	int x, y;
+	GdkColor color;
+	GtkWidget *widget;
+	GdkGC *gc;
+	guint32 shadow_color;
+	int shadow_offset;
+	
+	shadow_color = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (label),
+							   "eel-label-shadow-color"));
+	shadow_offset = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (label),
+							    "eel-label-shadow-offset"));
+
+	color = eel_gdk_rgb_to_color (shadow_color);
+	
+	get_layout_location (label, &x, &y);
+
+	widget = GTK_WIDGET (label);
+	if (shadow_offset > 0) {
+		gc = gdk_gc_new (widget->window);
+		gdk_gc_set_rgb_fg_color (gc, &color);
+		gdk_gc_set_clip_rectangle (gc, &event->area);
+		
+		gdk_draw_layout (widget->window,
+				 gc,
+				 x + shadow_offset, y + shadow_offset,
+				 label->layout);
+		g_object_unref (gc);
+	}
+	
+	gtk_paint_layout (widget->style,
+			  widget->window,
+			  GTK_WIDGET_STATE (widget),
+			  FALSE,
+			  &event->area,
+			  widget,
+			  "label",
+			  x, y,
+			  label->layout);
+
+	return TRUE;
+}
+
+static void
+eel_gtk_label_size_request (GtkLabel *label, GtkRequisition *requisition, gpointer user_data)
+{
+	gint shadow_offset;
+
+	shadow_offset = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (label),
+							    "eel-label-shadow-offset"));
+
+	requisition->width += shadow_offset;
+	requisition->height += shadow_offset;
+}
+
+static void
+set_up_label (GtkLabel *label)
+{
+
+	if (g_object_get_data (G_OBJECT (label), "eel-label-set-up") != NULL) {
+		return;
+	}
+
+	g_signal_connect (label, "expose_event",
+			  G_CALLBACK (eel_gtk_label_expose_event), NULL);
+	g_signal_connect_after (label, "size_request",
+				G_CALLBACK (eel_gtk_label_size_request), NULL);
+
+	g_object_set_data (G_OBJECT (label), "eel-label-set-up", "eel-label-set-up");
+}
+
+void
+eel_gtk_label_set_drop_shadow_color (GtkLabel *label,
+				     guint32 color)
+{
+	set_up_label (label);
+
+	g_object_set_data (G_OBJECT (label), "eel-label-shadow-color",
+			   GINT_TO_POINTER (color));
+		
+	gtk_widget_queue_draw (GTK_WIDGET (label));
+}
+
+void
+eel_gtk_label_set_drop_shadow_offset (GtkLabel *label,
+				      gint offset)
+{
+	set_up_label (label);
+
+	g_object_set_data (G_OBJECT (label), "eel-label-shadow-offset",
+			   GINT_TO_POINTER (offset));
+	
+	gtk_widget_queue_draw (GTK_WIDGET (label));
+}
+
 void
 eel_gtk_widget_set_background_color (GtkWidget *widget,
 				     const char *color_spec)
